# binary-tools

JavaScript binary parser for any browser or environment.  
Copyright (c) 2023  Rafael da Silva Rocha.  
https://rochars.com/binary-tools

**binary-tools** is a JavaScript binary parser for any browser or environment.

- **The fastest binary data parser for JavaScript**
- **Faster than TypedArrays and DataViews** for large datasets
- **Compatible with IE6+ and any environment with JavaScript support\***
- **Native support to 64-bit integers!\***
- **Zero dependencies!**
- **No polyfills!**
- **Native support to bfloat16 brain floating point format**
- **Native support to 16-bit half-precision floating point format**
- **Native support to 24-bit, 40-bit and 48-bit integers**
- **Support user-defined integer types (like Uint12)**
- **Support ASCII and ISO 8859-1 strings and characters**
- **Support UTF-8 strings**
- **Tailored for real-life use cases**
- **Less than 15kb minified!**


\* *64-bit integers packing/unpacking is only available on environments with native support to BigInt (ECMAScript 2020). The same goes for any user-defined integer type that is greater than 53-bit. You don't need to worry about this unless you are planning to pack/unpack numbers greater than 53-bit.*

## Tests

You can run the tests in your browser:  
[https://rochars.com/binary-tools/test/dist/browser](https://rochars.com/binary-tools/test/dist/browser)

The tests above use **mocha**, wich require some polyfills that may mask this module's working in older browsers. The link below runs *binary-tools.js* without loadig any file other than
*binary-tools.js* itself (only a few functions are executed, test suite is not used):  
[https://rochars.com/binary-tools/test/dist/es3](https://rochars.com/binary-tools/test/dist/es3)

This library is tested using **datasets with thousands of values generated with Python's struct module**. **8-bit and 16-bit integers are fully tested** with all possible values, signed and unsigned. Other types are tested with up to tens of thousands test values per type plus special cases (Infinity, -Infinity, NaN, overflows and so on).

## Install

Download it from [https://rochars.com/binary-tools](https://rochars.com/binary-tools). You can buy the source code and a MIT-Licensed version there as well. Other licenses are also available. All functions are available in all versions (including the free version).

You may also use a package manager:

```shell
npm install https://rochars.com/binary-tools-1.3.0.tgz
```

Or load it as a module from a CDN:

```javascript
import * as btools from "https://rochars.com/binary-tools@1.3.0/index.js";
```

My CDN is fast, safe and reliable. It is powered by the kind folks at [Cloudflare](https://cloudflare.com). Use it.

## Using in Node.js

The examples below assume you already installed *binary-tools* as a dependency in your project.

```javascript
const btools = require('binary-tools');

// Pack a signed 8-bit integer, returns a
// array with the number represented as bytes
let packed = btools.pack('h', -32765);
// packed == [3, 128]

let unpacked = btools.unpack('h', [255, 127]);
// unpacked = [32767]

// Pack a signed 16-bit integer to a existing byte buffer
// Start writing on index '4' of the buffer
let buffer = new Uint8Array(12);
btools.packTo('h', [1077], buffer, 4);
```

Or use ES6 imports if your software is a module:

```javascript
import * as btools from 'binary-tools';

let packed = btools.pack('h', -32765);
```

## Using in the Browser

Use the **binary-tools.js** file in the */dist* folder:

```html
<script src="./binary-tools/dist/binary-tools.js"></script>
<script>
  var packed = btools.pack('f', 2.1474836);
</script>
```

If you are targeting only modern browsers you may use the **./index.js** file in the root folder instead of *./dist/binary-tools.js*. The **./index.js** file have ES6-style exports.

### Browser compatibility

The file *./dist/binary-tools.js* is transpiled to ES3 and **is compatible with IE6+. It should work in all modern browsers** as well.

No polyfills are used or needed **unless you need to work with 64-bit numbers.** Packing and unpacking 64-bit numbers is only available for browsers that support ECMAScript 2020.

Cross-browser tests powered by Browserstack:  
<a href="https://www.browserstack.com"><img src="https://rochars.com/static/Browserstack-logo@2x.png" width="150px"/></a>

## Using in Deno / ES6 module

Use the **./index.js** file in the root folder:

```javascript
import * as btools from "./binary-tools/index.js";

let packed = btools.pack('h', -32765);
// packed == [3, 128]
```

Or load it from a CDN:

```javascript
import * as btools from "https://rochars.com/binary-tools@1.3.0/index.js";
```

---

## Quick Reference

### Difference of pack/unpack and write/read functions

The **packing** functions *pack()* and *packTo()* pack single values or structs, either returning an array with the values as bytes (in the case of *pack()*) or writing the packed values directly to a byte buffer (the case of *packTo()*).

Unpacking with *unpack()* or *unpackTo()* reads a single value or a struct from a buffer and return the values on an array (even if it is a single value) or write the results to a passed Array or TypedArray (in the case of *unpackTo*).

> **packing and unpacking functions can work with format strings made of many types (like 'hlhlhddd' or 'hlhlhtx4h')**.

---

The **writing** functions write one or many elements of a single type to a byte buffer (in the case of *writeTo()*) or return an array with the elements represented as bytes (in the case of *write()*).

Reading with *read()* or *readTo()* read one or many values of a single type from a byte buffer and return the values in an array (in the case of *read()*), or write the values directly to a TypedArray or Array (in the case of *writeTo()*).

> **reading and writing functions only work with a single type at a time (like 'h' or '>f')**. Any other type present in the format string after the first one will be ignored ('>Hhh' will be handled as '>H', for example). Byte order operators work the same they do in the packing/unpacking functions.

---

#### In summary:

If you are working with long sequences of values of the same type, **read/write** functions are **much faster**, and probably what you should use.

If you are working with file headers or other types of pre-defined structures of many different types, or single values, then **pack/unpack** functions are **more practical**, and probably what you should use.

They all share similar APIs.

---

### Standard types and format strings

Format strings specify the types being packed/unpacked and their byte order. Format strings for packing or unpacking structures may look like this:  

- 'HhHdH' 

where each character represent a type of data.

A format string may be preceded by a byte order operator. If no byte order operator is informed, little-endian is assumed. A format string with a byte order operator look like this:  

- '>HhHdH'  

where the > operator indicates big endian. Byte order operators may only appear at the start of the format string.

The byte order operators:
- '' - little endian (no operator)
- '<' - little endian
- '>' - big endian
- '!' - big endian

The standard type codes used for both packing/unpacking and reading/writing:
- 'x' - null pad byte, uses 1 byte
- 'c' - single ASCII character, uses 1 byte
- 'C' - single ISO-8859-1 character, uses 1 byte
- 's' - ASCII string, uses 1 byte per character
- 'S' - ISO-8859-1 string, uses 1 byte per character
- 'u' - UTF-8 string, may use 1 to 4 bytes per character
- 'b' - 8-bit signed integer, uses 1 byte
- 'B' - 8-bit unsigned integer, uses 1 byte
- '?' - boolean, uses 1 byte
- 'h' - 16-bit signed integer, uses 2 bytes
- 'H' - 16-bit unsigned integer, uses 2 bytes
- 'e' - 16-bit floating point, uses 2 bytes
- 'g' - 16-bit brain floating point, uses 2 bytes
- 't' - 24-bit signed integer, uses 3 bytes
- 'T' - 24-bit unsigned integer, uses 3 bytes
- 'l' - 32-bit signed integer, uses 4 bytes
- 'L' - 16-bit unsigned integer, uses 4 bytes
- 'f' - 32-bit floating point, uses 4 bytes
- 'j' - 40-bit signed integer, uses 5 bytes
- 'J' - 40-bit unsigned integer, uses 5 bytes
- 'k' - 48-bit signed integer, uses 6 bytes
- 'K' - 48-bit signed integer, uses 6 bytes
- 'd' - 64-bit floating point, uses 8 bytes
- 'q' - 64-bit signed integers, uses 8 bytes
- 'Q' - 64-bit unsigned integers, uses 8 bytes

A format character may be preceded by an integral repeat count. For example, the format string '4h' means exactly the same as 'hhhh'.

Notice that while this module use some names and conventions based on Python's struct module, it is not a JavaScript re-implementation of Python's struct module, neither intends to be. Some types (like 'c') work in a different way than they do in Python (or C).

#### Some considerations:
- there is no automatic padding. You must always format the padding according to your needs
- pad bytes 'x' are always packed as 0 and ignored when unpacking
- boolean '?' always pack values as either 0 or 1, and unpacks as either true or false
- using 0 as a repeat count in the format string is ignored ('0h' is the same as '1h' wich is the same as 'h').
- repeat counts placed at the end of the string will be ignored
- the types 'c' and 'C' represent single characters. A repeat count number for types 'c' and 'C' would represent the number of single characters in sequence - the format string '5c' on unpack [97,97,97,97,97] will return ['a','a','a','a','a'].
- the types 's' and 'S' represent strings.  A repeat count number for types 's' and 'S' would represent the number of characters in the string - the format string '5s' on unpack [97,97,97,97,97] will return 'aaaaa'.
- when packing, strings are padded with null bytes as appropriate to make it fit.
- Integers greater than 53-bit (like types q and Q) only work in environments that support BigInt.

All type codes map to objects like this:

```javascript
const int32 = {
  bits: 32, // number of bits used by the number, required
  signed: true, // optional, defaults to false
  fp: false // optional, defaults to false, true for floating-point numbers
}
```

You may define your own integer types:

```javascript
// Define a 12-bit signed integer
btools.types.z = { bits: 12, signed: true }
// Use your code just like you would use a default one:
btools.pack('z', 257);
```

Types with bits=0 and bits=1 are reserved for *pad bytes 'x'* and *booleans '?'*.

#### Byte order

To set the byte order to big endian, use the '>' operator:

```javascript
let packed = pack('>HH', [1, 1]);
// will return [0, 1, 0, 1]
```

To set to little endian, use the '<' operator:

```javascript
let packed = pack('<HH', [1, 1]);
// will return [1, 0, 1, 0]
```

The byte order operator may be ommited and will default to little-endian **regardless of the endianness of the host machine**:

```javascript
let packed = pack('H', 1);
// will return [1, 0] the same as pack(1, '<H');
```

Byte order operators placed anywhere except the start of the format string will be ignored. If more than one byte order operator is present, the first one from left to right is the one that will be used.

#### Pad bytes

Pad bytes (type 'x'), when present in the format string, dont need (and should not have) a matching element in the array of values. Pad bytes are packed as *0*. Pad bytes are only used with *packing* and *unpacking* functions. They should not be used with *writing* and *reading* functions.

---

### pack and packTo

**pack(format, values)** will return a *Array* with the bytes of the passed values. *values* can be a single item or an array.

```javascript
let packed = pack('h', 1000);
// return [232, 3]
packed = pack('>hh', [1000,1]);
//return [3, 232, 0, 1]
```

**packTo(format, values, buffer, index)** will write the bytes of the value to a buffer (any Array-like object). Writing starts on *index*. If no index is informed, it is assumed index=0.

```javascript
// Create a Uint8Array with size=4
let buffer = new Uint8Array(4);
// Start writing on index=2, writing the bytes of the 16-bit value to
// buffer[2] and buffer[3]. buffer[0] and buffer[1] are left untouched
packTo('h', 402, buffer, 2);
```

*index* can be ommited and will default to **zero**:

```javascript
// Create a Uint8Array with size=4
let buffer = new Uint8Array(4);
// Start writing on index=0, writing the bytes of the 16-bit value to
// buffer[0] and buffer[1]. buffer[2] and buffer[3] are left untouched
packTo('h', 402, buffer);
```

If the output buffer size is smaller than required by the data, it throws a
**Bad buffer length** error:

```javascript
// Create a Uint8Array with size=1
let buffer = new Uint8Array(1);
packTo('h', 402, buffer);
//Error: Bad buffer length
```

---

### write and writeTo

**write(format, values)** will return a *Array* with the bytes of the passed values.

```javascript
let packed = write('h', [1000, 1, 1, 1]);
// return [232, 3, 1, 0, 1, 0, 1, 0]
```

**writeTo(format, values, buffer, index)** will write the bytes of the value to the provided buffer (any Array-like object). Writing starts on *index*. If index is ommited, it is assumed index=0.

```javascript
// Create a Uint8Array with size=4
let buffer = new Uint8Array(4);
// Write all values to the buffer
writeTo('h', [402, 1], buffer);
```

*write()*, *writeTo*, *read()* and *readTo()* should be used with a single type at a time and are meant to read/write either large datasets or long sequences were all values are of the same type (such as in media files).

**read/write functions work with a single type at a time**. Any extra types defined in the format string after the first one will be ignored ('>Hbb' will be handled as '>H')

---

### Packing *null*, *false*, *true* and *undefined*

Attempts to pack or write **integers** with the following values:
- *undefined*
- *null*
- *true*
- *false*

will throw a **TypeError**.

If you wish, for example, to pack *1 for true* and *0 for false, null and undefined* you should either use the **boolean ('?')** type or change your input array. Notice that the *boolean* type will cause any value to become true or false when packing/unpacking.

---

### Unpacking and input buffer length

When reading values from a byte buffer **extra bytes in the end of the buffer are ignored** and **insufficient bytes will return a empty array** by default.

You can unpack in **safe mode** by setting the optional *safe* param to *true*. **In safe mode insufficient bytes in the input array or extra bytes in the end of the input array will cause a 'Bad buffer length' error**:

```javascript
// throws a 'Bad buffer length' error
let buffer = [0xff];
btools.readTo('H', buffer, output, 0, buffer.length, true);

// throws a 'Bad buffer length' error
buffer = [0xff, 0xff, 0xff];
btools.readTo('H', buffer, output, 0, buffer.length, true);

// throws a 'Bad buffer length' error
buffer = [0xff];
btools.unpack('H', buffer, 0, true);

// throws a 'Bad buffer length' error (start reading on index=2,
// attempt to unpack a 16-bit number from a single byte)
let buffer = [0xff, 0xff, 0xff];
btools.unpack('H', buffer, 2, true);

// throws a 'Bad buffer length' error (start reading on index=0,
// attempt to unpack a 16-bit number from 3 bytes)
buffer = [0xff, 0xff, 0xff];
btools.unpack('H', buffer, 0, true); 

// do not throw error (start reading on index=1,
// so skip the first byte and only read the last 2 bytes as a 16-bit number)
buffer = [0xff, 0xff, 0xff];
btools.unpack('H', buffer, 1, true); 
```

---

### Floating-point numbers

- Floating-point numbers are [IEEE 754](https://en.wikipedia.org/wiki/IEEE_754) standard.
- **Overflows** are rounded towards **Infinity** and **-Infinity**.
- **NaN** is packed as quiet NaN. Both **quiet NaN** and **signaling NaN** can be unpacked,
  both unpacked as *NaN*. Unpacking *NaN* with extra information on the significand is supported
  and will also result in *NaN* (extra information will be lost).
- Support packing and unpacking **negative zeros**.
- Support packing and unpacking **Infinity** and **negative Infinity**

#### Minifloats

Native support for 16-bit half-precision numbers (format code 'e') and for brain floating point numbers (format code 'g').

---

### Integers

- Overflow on integers will throw a *RangeError*.
- Packing values other than integers will throw a *TypeError*.
- You may *clamp* the input to avoid *RangeError* by setting **clamp** to **true**.

To clamp integers on overflow and avoid *RangeError*, set the optional *clamp* param to true:

```javascript
// Set clamp to true; values will be packed as their max or min values
// on overflow. In this case, packing an array of unsigned 8-bit ints
write('B', [1, 259, 2], true);
// will return [1, 255, 2]

// Set clamp to false; overflows cause a RangeError
write('B', [1, 259, 2], false);
// will throw a RangeError; this is the same as
write('B', [1, 259, 2]) // (omitting the clamp param),
// wich will also throw a RangeError
```

#### Signed integers

Signed integers are [two's complement](https://en.wikipedia.org/wiki/Two%27s_complement).

#### 64-bit integers

*binary-tools* have native support for packing and unpacking 64-bit integers on environments that support the BigInt class. Other types are not affected; if you are not using BigInts, you can use this module in any enviroment or browser with no polyfills or extra code needed.

Internally, all BigInts are represented using BigInt(), not *0n* notation.

---

### Characters and strings

Types 'c' and 'C' represent single characters. Type 'c' represent a single ASCII character (from 0 to 127). Type 'C' represent a single ISO-8859-1 character (from 0 to 255), covering all first 256 characters of Unicode. Both always use one byte.

Types 's' and 'S' represent strings. Type 's' represent a string of ASCII characters, and type 'S' represent a string of ISO-8859-1 characters. The strings use one byte per character.

Unless you need to enforce that the characters are 7-bit ASCII characters, you should generally use types 'C' and 'S'.

Single characters (types 'c' and 'C') behave the same way as the numeric types
when it comes to format strings:

```javascript
btools.pack('cH2c', ['a', 1, 'b', 'b']);
// will output [97, 1, 0, 98, 98]
```

For string types, the repeat count number represent the size of the string in bytes. Since every character uses exactly one byte, it is also the number of characters:

```javascript
btools.pack('sH2s', ['a', 1, 'bb']);
// will output [97, 1, 0, 98, 98]
```

If a string have less characters than the specified in the format, it will be padded with null bytes to fit the size defined in the format string:

```javascript
btools.pack('sH4s', ['a', 1, 'bb']);
// will output [97, 1, 0, 98, 98, 0, 0]
```

If a string have more characters than the specified in the format, it will be trimmed to fit the size defined in the format string:

```javascript
btools.pack('sH2s', ['a', 1, 'bbbb']);
// will output [97, 1, 0, 98, 98]
```

#### UTF-8 strings

For UTF-8 strings (type 'u') a repeat count number, if present, will represent the number of bytes used by string - but a character may use more than one byte. If there is no room for a character according to the given size (for example, the size is 3, but the character uses 4 bytes), the character will not be written and null bytes will be used to fit the size:

```javascript
// 輸 CJK COMPATIBILITY IDEOGRAPH-2F9DF (U+2F9DF) f0 af a7 9f
btools.pack('8u', '輸輸');
// will return [240, 175, 167, 159,   240, 175, 167, 159]

btools.pack('7u', '輸輸');
// will return [240, 175, 167, 159,   0, 0, 0]
```

If the size is greater than the string, then null bytes will be used to fit the size:
```javascript
// 輸 CJK COMPATIBILITY IDEOGRAPH-2F9DF (U+2F9DF) f0 af a7 9f
btools.pack('10u', '輸輸');
// will return [240, 175, 167, 159,   240, 175, 167, 159,   0, 0]
```

When unpacking UTF-8, if the last character is bigger than what the size allows it will be treated as a invalid character and the replacement character will be added to the string:

```javascript
// size dont reach the last byte of the last character
btools.unpack('7u', [240, 175, 167, 159,   240, 175, 167, 159];
// will return ['輸�']

// size match the 2 characters
btools.unpack('8u', [240, 175, 167, 159,   240, 175, 167, 159]
// will return ['輸輸']
```

When using type 'u' with *writing* functions like **write()**, if no start index or end index are given, then the full string will be packed:

```javascript
btools.write('u', '笠笠߹~$')
// will return
// [ 239, 167, 184,   239, 167, 184,   223, 185,   126,   36]
```

When using type 'u' with *reading* functions like **read()**, if no start index or end index are given, then the full string will be unpacked:

```javascript
btools.read('u', [239, 167, 184,   239, 167, 184,   223, 185,   126,   36])
// will return ['笠笠߹~$']
```

---

## API

### Packing/unpacking functions:

```javascript
/**
 * Pack a value or a struct to a byte buffer.
 * @param {string} format The struct format definition.
 * @param {!(TypedArray|Array<number>)} value The values to pack.
 * @param {!(Uint8Array|Array<number>)} buffer The byte buffer to write on.
 * @param {number=} [index=0] The buffer index to write.
 * @param {boolean=} [clamp=false] True to clamp ints on overflow.
 * @return {number} The next index to write.
 * @throws {Error} On unsupported type.
 * @throws {Error} If the output buffer size is not valid.
 * @throws {RangeError} On overflow if clamp is set to false.
 * @throws {TypeError} If 'value' is not a number or array of numbers.
 * @throws {TypeError} If 'value' is not all int and type is int.
 */
export function packTo(format, values, buffer, index=0, clamp=false) {}

/**
 * Unpack a value or a struct from an array of bytes to an Array-like object.
 * Always start writing the output at the beginning of the output array.
 * @param {string} format The struct format definition.
 * @param {!(Uint8Array|Array<number>)} buffer The byte buffer to unpack.
 * @param {!(TypedArray|Array<number>)} output The Array to write the values on.
 * @param {number=} [index=0] The buffer index to read.
 * @param {boolean=} [safe=false] If set to false, extra bytes in the end of
 *   the input array are ignored and input buffers with insufficient bytes will
 *   write nothing to the output array. If safe is set to true the function
 *   will throw a 'Bad buffer length' error on the aforementioned cases.
 * @throws {Error} On unsupported type.
 * @throws {Error} On bad input buffer length if on safe mode.
 */
export function unpackTo(format, buffer, output, index=0, safe=false) {}

/**
 * Pack a value or a struct as a array of bytes.
 * @param {string} format The struct format definition.
 * @param {!(TypedArray|Array<number>)} value The values to pack.
 * @param {boolean=} [clamp=false] True to clamp ints on overflow.
 * @return {!(Uint8Array|Array<number>)} The packed value.
 * @throws {Error} On unsupported type.
 * @throws {RangeError} On overflow if clamp is set to false.
 * @throws {TypeError} If 'value' is not a number.
 * @throws {TypeError} If 'value' is not a int and type is int.
 */
export function pack(format, value, clamp=false) {}

/**
 * Unpack a value or a struct from an array of bytes.
 * This method returns an Array even if only a single value is unpacked.
 * @param {string} format The struct format definition.
 * @param {!(Uint8Array|Array<number>)} buffer The byte buffer to unpack.
 * @param {number=} [index=0] The buffer index to read.
 * @param {boolean=} [safe=false] If set to false, extra bytes in the end of
 *   the input array are ignored and input buffers with insufficient bytes will
 *   write nothing to the output array. If safe is set to true the function
 *   will throw a 'Bad buffer length' error on the aforementioned cases.
 * @return {!(Array<number>|TypedArray)}
 * @throws {Error} On unsupported type.
 * @throws {Error} On bad input buffer length if on safe mode.
 */
export function unpack(format, buffer, index=0, safe=false) {}
```

### Writing/reading functions to handle long sequences of the same type:

```javascript
/**
 * Write a array of values to a byte buffer.
 * @param {string} format The format definition.
 * @param {!(TypedArray|Array<number>)} values The values to write.
 * @param {!(Uint8Array|Array<number>)} buffer The buffer to write on.
 * @param {number=} [index=0] The buffer index to start writing.
 * @param {boolean=} [clamp=false] True to clamp ints on overflow.
 * @return {number} The next index to write on the buffer.
 * @throws {Error} On unsupported type.
 * @throws {RangeError} On integer overflow if clamp is set to false.
 * @throws {TypeError} If 'values' is not an array of numbers.
 * @throws {TypeError} If 'values' is not an array of ints and type is int.
 */
export function writeTo(format, values, buffer, index=0, clamp=false) {}

/**
 * Unpack a array of values from a byte buffer to a array or a typed array.
 * @param {string} format The format definition.
 * @param {!(Uint8Array|Array<number>)} buffer The byte buffer.
 * @param {!(TypedArray|Array<number>)} output The output array.
 * @param {number=} [start=0] The input buffer index to start reading.
 * @param {number=} [end=buffer.length] The input buffer index to stop reading.
 * @param {boolean=} [safe=false] If set to false, extra bytes in the end of
 *   the input array are ignored and input buffers with insufficient bytes will
 *   write nothing to the output array. If safe is set to true the function
 *   will throw a 'Bad buffer length' error on the aforementioned cases.
 * @throws {Error} On unsupported type.
 * @throws {Error} On bad input buffer length if on safe mode.
 */
export function readTo(
    format, buffer, output, start=0, end=buffer.length, safe=false) {}

/**
 * Write a array of values as a array of bytes.
 * @param {string} format The format definition.
 * @param {!(TypedArray|Array<number>)} values The values to pack.
 * @param {boolean=} [clamp=false] True to clamp ints on overflow.
 * @return {!Array<number>} The packed values.
 * @throws {Error} On unsupported type.
 * @throws {RangeError} On overflow if clamp is set to false.
 * @throws {TypeError} If 'values' is not a array of numbers.
 * @throws {TypeError} If 'values' is not a array of ints and type is int.
 */
export function write(format, values, clamp=false) {}

/**
 * Read a array of values from a byte buffer.
 * @param {string} format The format definition.
 * @param {!(Uint8Array|Array<number>)} buffer The byte buffer.
 * @param {number=} [start=0] The buffer index to start reading.
 * @param {number=} [end=buffer.length] The buffer index to stop reading.
 * @param {boolean=} [safe=false] If set to false, extra bytes in the end of
 *   the input array are ignored and input buffers with insufficient bytes will
 *   write nothing to the output array. If safe is set to true the function
 *   will throw a 'Bad buffer length' error on the aforementioned cases.
 * @return {!Array<number>}
 * @throws {Error} On unsupported type.
 * @throws {Error} On bad input buffer length if on safe mode.
 */
export function read(
    format, buffer, start=0, end=buffer.length, safe=false) {}
```

Note that in version 1.x the *end* param for **read** and **readTo** is non-inclusive, so it must be set always as index + 1. For example, to read from array position 0 to position 7 you should *read('u', buffer, 0, 8)*.

### Tools:

```javascript
/**
 * Swap the byte ordering in a buffer. The buffer is modified in place.
 * @param {!(Uint8Array|Array<number>)} buffer The bytes.
 * @param {number} offset The byte offset.
 * @param {number=} [start=0] The start index.
 * @param {number=} [end=bytes.length] The end index.
 */
export function endianness(buffer, offset, start=0, end=buffer.length) {}

/**
 * Calculate the buffer size based on a format string.
 * @param {string} format The format string.
 * @param {boolean=} [includePads=false] True if padding bytes should be counted.
 * @return {number}
 * @throws {Error} On unsupported type.
 */
export function calcSize(format, includePads=false) {}
```

Note that in future versions the *includePads* param for **calcSize** will be set to *true* by default. This param is kept like this in 1.x releases as to not break the 1.x interface. Apart from a few situations, you should always use calcSize with includePads set to *true*.

Note that in version 1.x the *end* param for **endiannness** is non-inclusive, so it must be set always as index + 1.

### The **types** object:

```javascript
/**
 * A binary-tools datatype definition object.
 * @typedef {Object} DataType
 * @property {number} bits The number of bits used by the type.
 * @property {boolean} signed False for unsigned, true for signed.
 * @property {boolean} fp False for integers, true for floating point numbers.
 * @property {boolean} b False for all except formats c, C, s, S and g.
 * @property {boolean} v False for all except formats s, S and u.
 */

/**
 * The standard binary-tools datatype definitions.
 * Types can be created by adding new properties named as
 * a single character to this object in the format:
 * { bits: number, signed: boolean, fp: boolean }
 * Numbers and the characters @=<>! are reserved, and also
 * all characters used for pre-defined types (xcCsSubB?hHegtTlLfjJkKd).
 * @enum {Object}
 * @property {DataType} x Pad byte
 * @property {DataType} c ASCII character
 * @property {DataType} C Extended ASCII (ISO-8859-1) character
 * @property {DataType} s ASCII string
 * @property {DataType} S Extended ASCII (ISO-8859-1) string
 * @property {DataType} u UTF-8 string
 * @property {DataType} b Signed 8-bit integers
 * @property {DataType} B Unsigned 8-bit integers
 * @property {DataType} ? Boolean
 * @property {DataType} h Signed 16-bit integers
 * @property {DataType} H Unsigned 16-bit integers
 * @property {DataType} e 16-bit Half-precision floating-point numbers
 * @property {DataType} g 16-bit Brain floating point numbers
 * @property {DataType} t Signed 24-bit integers
 * @property {DataType} T Unsigned 24-bit integers
 * @property {DataType} l Signed 32-bit integers
 * @property {DataType} L Unsigned 32-bit integers
 * @property {DataType} f 32-bit Single-precision floating-point numbers
 * @property {DataType} j Signed 40-bit integers
 * @property {DataType} J Unsigned 40-bit integers
 * @property {DataType} k Signed 48-bit integers
 * @property {DataType} K Unsigned 48-bit integers
 * @property {DataType} d 64-bit Double-precision floating-point numbers
 * @property {DataType} q signed 64-bit integers
 * @property {DataType} Q Unsigned 64-bit integers
 * @memberOf module:binary-tools
 */
export const types = {
  x: {bits: 0, signed: false, fp: false, b: false, v: false},
  c: {bits: 7, signed: false, fp: false, b: true, v: false},
  C: {bits: 8, signed: false, fp: false, b: true, v: false},
  s: {bits: 7, signed: false, fp: false, b: true, v: true},
  S: {bits: 8, signed: false, fp: false, b: true, v: true},
  u: {bits: undefined, signed: false, fp: false, b: false, v: true},
  b: {bits: 8, signed: true, fp: false, b: false, v: false},
  B: {bits: 8, signed: false, fp: false, b: false, v: false},
  '?': {bits: 1, signed: false, fp: false, b: false, v: false},
  h: {bits: 16, signed: true, fp: false, b: false, v: false},
  H: {bits: 16, signed: false, fp: false, b: false, v: false},
  e: {bits: 16, signed: true, fp: true, b: false, v: false},
  g: {bits: 16, signed: true, fp: true, b: true, v: false},
  t: {bits: 24, signed: true, fp: false, b: false, v: false},
  T: {bits: 24, signed: false, fp: false, b: false, v: false},
  l: {bits: 32, signed: true, fp: false, b: false, v: false},
  L: {bits: 32, signed: false, fp: false, b: false, v: false},
  f: {bits: 32, signed: true, fp: true, b: false, v: false},
  j: {bits: 40, signed: true, fp: false, b: false, v: false},
  J: {bits: 40, signed: false, fp: false, b: false, v: false},
  k: {bits: 48, signed: true, fp: false, b: false, v: false},
  K: {bits: 48, signed: false, fp: false, b: false, v: false},
  d: {bits: 64, signed: true, fp: true, b: false, v: false},
  q: {bits: 64, signed: true, fp: false, b: false, v: false},
  Q: {bits: 64, signed: false, fp: false, b: false, v: false}
};
```

## Building

Information in this section is only relevant if you are developing *binary-tools.js* itself or need to expose *binary-tools.js* methods on your software's API.

### Tests on big-endian systems

Use [QEMU](https://www.qemu.org/) with this PowerPC/Debian image:  
https://people.debian.org/~aurel32/qemu/powerpc/

### Compilation

**binary-tools** dist files are created with [Closure Compiler](https://github.com/google/closure-compiler-js) with compilation level set to ADVANCED, so properties that have not been exported will be renamed (and likely result in errors). During the build process the tests are run against both source files and the compiled dist files to ensure all went OK during the compilation.

### Style guide

**binary-tools** source code should follow the Google JavaScript Style Guide:  
https://google.github.io/styleguide/jsguide.html

---

## Additional files

The TypeScript declarations are in *./dist/binary-tools.d.ts*

The Closure Compiler externs are in *./externs/binary-tools.js*

---

## Reporting issues

Report security issues to this e-mail: rocha.rafaelsilva@gmail.com.

---

## Buying a fully MIT-Licensed version of this software:

You can buy the MIT-licensed (aka Expat License) version of this software at [https://rochars.com/binary-tools](https://rochars.com/binary-tools) and use it as you please according to the terms of the MIT License. Other licensing options are available, too. They include the complete source code, the tests, documentation and all files related to the project, all released under more permissive licenses (MIT, BSD, Apache, GPL and so) for your convenience.

The unpaid version of this software is released under the CC-BY-NC-ND-4.0 License. You are free to use it under [the terms of the CC-BY-NC-ND-4.0 License](href="https://creativecommons.org/licenses/by-nc-nd/4.0/us/). View [COPYING](https://rochars.com/binary-tools/COPYING) for more information.

---

## LICENSE

binary-tools: JavaScript binary parser for any browser or environment.  
Copyright (c) 2023 Rafael da Silva Rocha

See LICENSE, NOTICE and COPYING for more information.
