// Type definitions for binary-tools 1.3
// Project: https://rochars.com/binary-tools
// Definitions by: Rafael da Silva Rocha <rocha.rafaelsilva@gmail.com>

// packing/unpacking

/**
 * Pack a value or a struct to a byte buffer.
 * @param {string} format The struct format definition.
 * @param {!(TypedArray|Array<number>)} value The values to pack.
 * @param {!(Uint8Array|Array<number>)} buffer The byte buffer to write on.
 * @param {number=} [index=0] The buffer index to write.
 * @param {boolean=} [clamp=false] True to clamp ints on overflow.
 * @return {number} The next index to write.
 * @throws {Error} On unsupported type.
 * @throws {Error} If the output buffer size is not valid.
 * @throws {RangeError} On overflow if clamp is set to false.
 * @throws {TypeError} If 'value' is not a number or array of numbers.
 * @throws {TypeError} If 'value' is not all int and type is int.
 */
export function packTo(
  theType: string,
  value: number|ArrayLike<number>,
  buffer: ArrayLike<number>,
  index?: number,
  clamp?: boolean): number;

/**
 * Unpack a value or a struct from an array of bytes to an Array-like object.
 * Always start writing the output at the beginning of the output array.
 * @param {string} format The struct format definition.
 * @param {!(Uint8Array|Array<number>)} buffer The byte buffer to unpack.
 * @param {!(TypedArray|Array<number>)} output The Array to write the values on.
 * @param {number=} [index=0] The buffer index to read.
 * @param {boolean=} [safe=false] If set to false, extra bytes in the end of
 *   the input array are ignored and input buffers with insufficient bytes will
 *   write nothing to the output array. If safe is set to true the function
 *   will throw a 'Bad buffer length' error on the aforementioned cases.
 * @throws {Error} On unsupported type.
 * @throws {Error} On bad input buffer length if on safe mode.
 */
export function unpackTo(
  format: string,
  buffer: ArrayLike<number>,
  output: ArrayLike<number>,
  index?: number,
  safe?: boolean): void;

/**
 * Pack a value or a struct as a array of bytes.
 * @param {string} format The struct format definition.
 * @param {!(TypedArray|Array<number>)} value The values to pack.
 * @param {boolean=} [clamp=false] True to clamp ints on overflow.
 * @return {!(Uint8Array|Array<number>)} The packed value.
 * @throws {Error} On unsupported type.
 * @throws {RangeError} On overflow if clamp is set to false.
 * @throws {TypeError} If 'value' is not a number.
 * @throws {TypeError} If 'value' is not a int and type is int.
 */
export function pack(
  theType: string,
  value: number|ArrayLike<number>,
  clamp?: boolean): Array<number>;

/**
 * Unpack a value or a struct from an array of bytes.
 * This method returns an Array even if only a single value is unpacked.
 * @param {string} format The struct format definition.
 * @param {!(Uint8Array|Array<number>)} buffer The byte buffer to unpack.
 * @param {number=} [index=0] The buffer index to read.
 * @param {boolean=} [safe=false] If set to false, extra bytes in the end of
 *   the input array are ignored and input buffers with insufficient bytes will
 *   write nothing to the output array. If safe is set to true the function
 *   will throw a 'Bad buffer length' error on the aforementioned cases.
 * @return {!(Array<number>|TypedArray)}
 * @throws {Error} On unsupported type.
 * @throws {Error} On bad input buffer length if on safe mode.
 */
export function unpack(
  theType: string,
  buffer: ArrayLike<number>,
  index?: number,
  safe?: boolean): Array<number>;

// reading/writing

/**
 * Write a array of values to a byte buffer.
 * @param {string} format The format definition.
 * @param {!(TypedArray|Array<number>)} values The values to write.
 * @param {!(Uint8Array|Array<number>)} buffer The buffer to write on.
 * @param {number=} [index=0] The buffer index to start writing.
 * @param {boolean=} [clamp=false] True to clamp ints on overflow.
 * @return {number} The next index to write on the buffer.
 * @throws {Error} On unsupported type.
 * @throws {RangeError} On integer overflow if clamp is set to false.
 * @throws {TypeError} If 'values' is not an array of numbers.
 * @throws {TypeError} If 'values' is not an array of ints and type is int.
 */
export function writeTo(
  theType: string,
  values: number|ArrayLike<number>,
  buffer: ArrayLike<number>,
  index?: number,
  clamp?: boolean): number;

/**
 * Unpack a array of values from a byte buffer to a array or a typed array.
 * @param {string} format The format definition.
 * @param {!(Uint8Array|Array<number>)} buffer The byte buffer.
 * @param {!(TypedArray|Array<number>)} output The output array.
 * @param {number=} [start=0] The input buffer index to start reading.
 * @param {number=} [end=buffer.length] The input buffer index to stop reading.
 * @param {boolean=} [safe=false] If set to false, extra bytes in the end of
 *   the input array are ignored and input buffers with insufficient bytes will
 *   write nothing to the output array. If safe is set to true the function
 *   will throw a 'Bad buffer length' error on the aforementioned cases.
 * @throws {Error} On unsupported type.
 * @throws {Error} On bad input buffer length if on safe mode.
 */
export function readTo(
  theType: string,
  buffer: ArrayLike<number>,
  output: ArrayLike<number>,
  start?: number,
  end?: number,
  safe?: boolean): void;

/**
 * Write a array of values as a array of bytes.
 * @param {string} format The format definition.
 * @param {!(TypedArray|Array<number>)} values The values to pack.
 * @param {boolean=} [clamp=false] True to clamp ints on overflow.
 * @return {!Array<number>} The packed values.
 * @throws {Error} On unsupported type.
 * @throws {RangeError} On overflow if clamp is set to false.
 * @throws {TypeError} If 'values' is not a array of numbers.
 * @throws {TypeError} If 'values' is not a array of ints and type is int.
 */
export function write(
  theType: string,
  values: number|ArrayLike<number>,
  clamp?:boolean): Array<number>;

/**
 * Read a array of values from a byte buffer.
 * @param {string} format The format definition.
 * @param {!(Uint8Array|Array<number>)} buffer The byte buffer.
 * @param {number=} [start=0] The buffer index to start reading.
 * @param {number=} [end=buffer.length] The buffer index to stop reading.
 * @param {boolean=} [safe=false] If set to false, extra bytes in the end of
 *   the input array are ignored and input buffers with insufficient bytes will
 *   write nothing to the output array. If safe is set to true the function
 *   will throw a 'Bad buffer length' error on the aforementioned cases.
 * @return {!Array<number>}
 * @throws {Error} On unsupported type.
 * @throws {Error} On bad input buffer length if on safe mode.
 */
export function read(
  theType: string,
  buffer: ArrayLike<number>,
  start?: number,
  end?: number,
  safe?: boolean): Array<number>;

// tools

/**
 * Swap the byte ordering in a buffer. The buffer is modified in place.
 * @param {!(Uint8Array|Array<number>)} bytes The bytes.
 * @param {number} offset The byte offset.
 * @param {number=} [start=0] The start index.
 * @param {number=} [end=bytes.length] The end index.
 */
export function endianness(
  bytes: ArrayLike<number>,
  offset: number,
  start?: number,
  end?: number): void;

/**
 * Calculate the buffer size based on a format string.
 * @param {string} format The format string.
 * @param {boolean=} [includePads=false] True if padding bytes should be counted.
 * @return {number}
 * @throws {Error} On unsupported type.
 */
export function calcSize(
  format: string,
  includePads?: boolean): number;

// types

/**
 * A binary-tools datatype definition object.
 * @typedef {Object} DataType
 * @property {number} bits The number of bits used by the type.
 * @property {boolean} signed False for unsigned, true for signed.
 * @property {boolean} fp False for integers, true for floating point numbers.
 * @property {boolean} b False for all except formats c, C, s, S and g.
 * @property {boolean} v False for all except formats s, S and u.
 */

/**
 * The standard binary-tools datatype definitions.
 * Types can be created by adding new properties named as
 * a single character to this object in the format:
 * { bits: number, signed: boolean, fp: boolean }
 * Numbers and the characters @=<>! are reserved, and also
 * all characters used for pre-defined types (xcCsSubB?hHegtTlLfjJkKd).
 * @enum {Object}
 * @property {DataType} x Pad byte
 * @property {DataType} c ASCII character
 * @property {DataType} C Extended ASCII (ISO-8859-1) character
 * @property {DataType} s ASCII string
 * @property {DataType} S Extended ASCII (ISO-8859-1) string
 * @property {DataType} u UTF-8 string
 * @property {DataType} b Signed 8-bit integers
 * @property {DataType} B Unsigned 8-bit integers
 * @property {DataType} ? Boolean
 * @property {DataType} h Signed 16-bit integers
 * @property {DataType} H Unsigned 16-bit integers
 * @property {DataType} e 16-bit Half-precision floating-point numbers
 * @property {DataType} g 16-bit Brain floating point numbers
 * @property {DataType} t Signed 24-bit integers
 * @property {DataType} T Unsigned 24-bit integers
 * @property {DataType} l Signed 32-bit integers
 * @property {DataType} L Unsigned 32-bit integers
 * @property {DataType} f 32-bit Single-precision floating-point numbers
 * @property {DataType} j Signed 40-bit integers
 * @property {DataType} J Unsigned 40-bit integers
 * @property {DataType} k Signed 48-bit integers
 * @property {DataType} K Unsigned 48-bit integers
 * @property {DataType} d 64-bit Double-precision floating-point numbers
 * @property {DataType} q signed 64-bit integers
 * @property {DataType} Q Unsigned 64-bit integers
 * @memberOf module:binary-tools
 */
export let types;
