/**
 * @fileoverview Externs for binary-tools 1.3
 * @see {@link https://rochars.com/binary-tools}
 * @externs
 * @license
 * Copyright (c) 2023 Rafael da Silva Rocha.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

/** @type {!Object} */
var btools = {};

// packing/unpacking

/**
 * Pack a value or a struct to a byte buffer.
 * @param {string} format The struct format definition.
 * @param {!(TypedArray|Array<number>)} value The values to pack.
 * @param {!(Uint8Array|Array<number>)} buffer The byte buffer to write on.
 * @param {number=} [index=0] The buffer index to write.
 * @param {boolean=} [clamp=false] True to clamp ints on overflow.
 * @return {number} The next index to write.
 * @throws {Error} On unsupported type.
 * @throws {Error} If the output buffer size is not valid.
 * @throws {RangeError} On overflow if clamp is set to false.
 * @throws {TypeError} If 'value' is not a number or array of numbers.
 * @throws {TypeError} If 'value' is not all int and type is int.
 */
btools.packTo = function(format, value, buffer, index=0, clamp=false) {};

/**
 * Unpack a value or a struct from an array of bytes to an Array-like object.
 * Always start writing the output at the beginning of the output array.
 * @param {string} format The struct format definition.
 * @param {!(Uint8Array|Array<number>)} buffer The byte buffer to unpack.
 * @param {!(TypedArray|Array<number>)} output The Array to write the values on.
 * @param {number=} [index=0] The buffer index to read.
 * @param {boolean=} [safe=false] If set to false, extra bytes in the end of
 *   the input array are ignored and input buffers with insufficient bytes will
 *   write nothing to the output array. If safe is set to true the function
 *   will throw a 'Bad buffer length' error on the aforementioned cases.
 * @throws {Error} On unsupported type.
 * @throws {Error} On bad input buffer length if on safe mode.
 */
btools.unpackTo = function(format, buffer, output, index=0, safe=false) {};

/**
 * Pack a value or a struct as a array of bytes.
 * @param {string} format The struct format definition.
 * @param {!(TypedArray|Array<number>)} value The values to pack.
 * @param {boolean=} [clamp=false] True to clamp ints on overflow.
 * @return {!(Uint8Array|Array<number>)} The packed value.
 * @throws {Error} On unsupported type.
 * @throws {RangeError} On overflow if clamp is set to false.
 * @throws {TypeError} If 'value' is not a number.
 * @throws {TypeError} If 'value' is not a int and type is int.
 */
btools.pack = function(format, value, clamp=false) {};

/**
 * Unpack a value or a struct from an array of bytes.
 * This method returns an Array even if only a single value is unpacked.
 * @param {string} format The struct format definition.
 * @param {!(Uint8Array|Array<number>)} buffer The byte buffer to unpack.
 * @param {number=} [index=0] The buffer index to read.
 * @param {boolean=} [safe=false] If set to false, extra bytes in the end of
 *   the input array are ignored and input buffers with insufficient bytes will
 *   write nothing to the output array. If safe is set to true the function
 *   will throw a 'Bad buffer length' error on the aforementioned cases.
 * @return {!(Array<number>|TypedArray)}
 * @throws {Error} On unsupported type.
 * @throws {Error} On bad input buffer length if on safe mode.
 */
btools.unpack = function(format, buffer, index=0, safe=false) {};

// writing/reading

/**
 * Write a array of values to a byte buffer.
 * @param {string} format The format definition.
 * @param {!(TypedArray|Array<number>)} values The values to write.
 * @param {!(Uint8Array|Array<number>)} buffer The buffer to write on.
 * @param {number=} [index=0] The buffer index to start writing.
 * @param {boolean=} [clamp=false] True to clamp ints on overflow.
 * @return {number} The next index to write on the buffer.
 * @throws {Error} On unsupported type.
 * @throws {RangeError} On integer overflow if clamp is set to false.
 * @throws {TypeError} If 'values' is not an array of numbers.
 * @throws {TypeError} If 'values' is not an array of ints and type is int.
 */
btools.writeTo = function(format, values, buffer, index=0,
    clamp=false) {};

/**
 * Unpack a array of values from a byte buffer to a array or a typed array.
 * @param {string} format The format definition.
 * @param {!(Uint8Array|Array<number>)} buffer The byte buffer.
 * @param {!(TypedArray|Array<number>)} output The output array.
 * @param {number=} [start=0] The input buffer index to start reading.
 * @param {number=} [end=buffer.length] The input buffer index to stop reading.
 * @param {boolean=} [safe=false] If set to false, extra bytes in the end of
 *   the input array are ignored and input buffers with insufficient bytes will
 *   write nothing to the output array. If safe is set to true the function
 *   will throw a 'Bad buffer length' error on the aforementioned cases.
 * @throws {Error} On unsupported type.
 * @throws {Error} On bad input buffer length if on safe mode.
 */
btools.readTo = function(
    format, buffer, output, start=0, end=buffer.length, safe=false) {};

/**
 * Write a array of values as a array of bytes.
 * @param {string} format The format definition.
 * @param {!(TypedArray|Array<number>)} values The values to pack.
 * @param {boolean=} [clamp=false] True to clamp ints on overflow.
 * @return {!Array<number>} The packed values.
 * @throws {Error} On unsupported type.
 * @throws {RangeError} On overflow if clamp is set to false.
 * @throws {TypeError} If 'values' is not a array of numbers.
 * @throws {TypeError} If 'values' is not a array of ints and type is int.
 */
btools.write = function(format, values, clamp=false) {};

/**
 * Read a array of values from a byte buffer.
 * @param {string} format The format definition.
 * @param {!(Uint8Array|Array<number>)} buffer The byte buffer.
 * @param {number=} [start=0] The buffer index to start reading.
 * @param {number=} [end=buffer.length] The buffer index to stop reading.
 * @param {boolean=} [safe=false] If set to false, extra bytes in the end of
 *   the input array are ignored and input buffers with insufficient bytes will
 *   write nothing to the output array. If safe is set to true the function
 *   will throw a 'Bad buffer length' error on the aforementioned cases.
 * @return {!Array<number>}
 * @throws {Error} On unsupported type.
 * @throws {Error} On bad input buffer length if on safe mode.
 */
btools.read = function (
  format, buffer, start=0, end=buffer.length, safe=false) {};

// tools

/**
 * Swap the byte ordering in a buffer. The buffer is modified in place.
 * @param {!(Uint8Array|Array<number>)} bytes The bytes.
 * @param {number} offset The byte offset.
 * @param {number=} [start=0] The start index.
 * @param {number=} [end=bytes.length] The end index.
 */
btools.endianness = function(bytes, offset, start=0, end=bytes.length) {};

/**
 * Calculate the buffer size based on a format string.
 * @param {string} format The format string.
 * @param {boolean=} [includePads=false] True if padding bytes should be counted.
 * @return {number}
 * @throws {Error} On unsupported type.
 */
btools.calcSize = function(format, includePads=false) {};

// types

/**
 * A binary-tools datatype definition object.
 * @typedef {Object} DataType
 * @property {number} bits The number of bits used by the type.
 * @property {boolean} signed False for unsigned, true for signed.
 * @property {boolean} fp False for integers, true for floating point numbers.
 * @property {boolean} b False for all except formats c, C, s, S and g.
 * @property {boolean} v False for all except formats s, S and u.
 */

/**
 * The standard binary-tools datatype definitions.
 * Types can be created by adding new properties named as
 * a single character to this object in the format:
 * { bits: number, signed: boolean, fp: boolean }
 * Numbers and the characters @=<>! are reserved, and also
 * all characters used for pre-defined types (xcCsSubB?hHegtTlLfjJkKd).
 * @enum {Object}
 * @property {DataType} x Pad byte
 * @property {DataType} c ASCII character
 * @property {DataType} C Extended ASCII (ISO-8859-1) character
 * @property {DataType} s ASCII string
 * @property {DataType} S Extended ASCII (ISO-8859-1) string
 * @property {DataType} u UTF-8 string
 * @property {DataType} b Signed 8-bit integers
 * @property {DataType} B Unsigned 8-bit integers
 * @property {DataType} ? Boolean
 * @property {DataType} h Signed 16-bit integers
 * @property {DataType} H Unsigned 16-bit integers
 * @property {DataType} e 16-bit Half-precision floating-point numbers
 * @property {DataType} g 16-bit Brain floating point numbers
 * @property {DataType} t Signed 24-bit integers
 * @property {DataType} T Unsigned 24-bit integers
 * @property {DataType} l Signed 32-bit integers
 * @property {DataType} L Unsigned 32-bit integers
 * @property {DataType} f 32-bit Single-precision floating-point numbers
 * @property {DataType} j Signed 40-bit integers
 * @property {DataType} J Unsigned 40-bit integers
 * @property {DataType} k Signed 48-bit integers
 * @property {DataType} K Unsigned 48-bit integers
 * @property {DataType} d 64-bit Double-precision floating-point numbers
 * @property {DataType} q signed 64-bit integers
 * @property {DataType} Q Unsigned 64-bit integers
 * @memberOf module:binary-tools
 */
btools.types = {
  x: {bits: 0, signed: false, fp: false, b: false, v: false},
  c: {bits: 7, signed: false, fp: false, b: true, v: false},
  C: {bits: 8, signed: false, fp: false, b: true, v: false},
  s: {bits: 7, signed: false, fp: false, b: true, v: true},
  S: {bits: 8, signed: false, fp: false, b: true, v: true},
  u: {signed: false, fp: false, b: false, v: true},
  b: {bits: 8, signed: true, fp: false, b: false, v: false},
  B: {bits: 8, signed: false, fp: false, b: false, v: false},
  '?': {bits: 1, signed: false, fp: false, b: false, v: false},
  h: {bits: 16, signed: true, fp: false, b: false, v: false},
  H: {bits: 16, signed: false, fp: false, b: false, v: false},
  e: {bits: 16, signed: true, fp: true, b: false, v: false},
  g: {bits: 16, signed: true, fp: true, b: true, v: false},
  t: {bits: 24, signed: true, fp: false, b: false, v: false},
  T: {bits: 24, signed: false, fp: false, b: false, v: false},
  l: {bits: 32, signed: true, fp: false, b: false, v: false},
  L: {bits: 32, signed: false, fp: false, b: false, v: false},
  f: {bits: 32, signed: true, fp: true, b: false, v: false},
  j: {bits: 40, signed: true, fp: false, b: false, v: false},
  J: {bits: 40, signed: false, fp: false, b: false, v: false},
  k: {bits: 48, signed: true, fp: false, b: false, v: false},
  K: {bits: 48, signed: false, fp: false, b: false, v: false},
  d: {bits: 64, signed: true, fp: true, b: false, v: false},
  q: {bits: 64, signed: true, fp: false, b: false, v: false},
  Q: {bits: 64, signed: false, fp: false, b: false, v: false}
};
